#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <assert.h>

#include <string.h>
#include <ctype.h>

#include "parser.h"
#include "resmap.h"
#include "posdb.h"

static TResMap GlobalMap;

static int ListResources, ListOptions, ListIDs, ListSyms;

/*--------------------------------------------------------------------------*/
/* parser help routines */

static unsigned int GetMask(char *pAsc, unsigned int *pmask)
{
  int l = strlen(pAsc) - 1, base;
  unsigned int res;
  static char *digvals = "0123456789ABCDEF";

  if (toupper(pAsc[l]) == 'H')
  {
    pAsc[l] = '\0';
    base = 16;
  }
  else if (toupper(pAsc[l]) == 'B')
  {
    pAsc[l] = '\0';
    base = 2;
  }
  else
    base = 8;

  res = *pmask = 0;
  while (*pAsc)
  {
    res *= base; *pmask *= base;
    if (toupper(*pAsc) == 'X')
      *pmask |= (base - 1);
    else
      res += strchr(digvals, toupper(*pAsc)) - digvals;
    pAsc++;
  }

  *pmask = ~(*pmask);
  return res;
}

static FILE *GetADF(unsigned int num, char Prefix)
{
  char Name[20];
  FILE *f;

  sprintf(Name, "%c%04x.adf", Prefix, num);
  if ((f = fopen(Name, "r")))
    return f;

  sprintf(Name, "%c%04X.adf", toupper(Prefix), num);
  if ((f = fopen(Name, "r")))
    return f;

  sprintf(Name, "%c%04x.ADF", Prefix, num);
  if ((f = fopen(Name, "r")))
    return f;

  sprintf(Name, "%c%04X.ADF", toupper(Prefix), num);
  if ((f = fopen(Name, "r")))
    return f;

  return NULL;
}

/*--------------------------------------------------------------------------*/
/* main routine(s) */

void ChkNameFound(int *NameFound, char prefix, int posid)
{
  if (!(*NameFound))
  {
    *NameFound = 1;
    switch (toupper(prefix))
    {
      case 'D':
        switch (posid)
        {
          case 0xdefe:
            printf("IBM PS/2 Model 77 'Bermuda' Planar\n"); break;
          case 0xdffe:
            printf("IBM PS/2 Model 76 'Bermuda' Planar\n"); break;
          case 0xe0fe:
            printf("IBM PS/2 Model 57 (9557) Planar\n"); break;
          case 0xe1fe:
            printf("IBM PS/2 Model 56 (9556) Planar\n"); break;
          case 0xe2fe:
            printf("IBM PS/2 Model 57 (8557) Planar\n"); break;
          case 0xe3ff:
            printf("IBM PS/2 Model 65 Planar\n"); break;
          case 0xe4fe:
            printf("IBM Server 95 / Server 500 Planar\n"); break;
          case 0xe6fe:
            printf("IBM PS/2 Model 56 (8556) Planar\n"); break;
          case 0xee7a:
            printf("IBM PS/2 Model 77(I/S) 'Lacuna' Planar\n"); break;
          case 0xee7c:
            printf("IBM PS/2 Model 76(I/S) 'Lacuna' Planar\n"); break;
          case 0xff6f:
            printf("IBM PS/2 Model 90 (8590/9590) Planar\n"); break;
          case 0xff9f:
            printf("IBM PS/2 Model 95 (8595/9595) Planar\n"); break;
          case 0xffaf:
            printf("IBM PS/2 Model P70/386 Planar\n"); break;
          case 0xfff9:
            printf("IBM PS/2 Model 80 Planar\n"); break;
          default:
            printf("unknown planar ID %04x\n", posid);
        }
        break;
      default:
        printf("unknown adapter ID %04x\n", posid);
    }
  }
}

static void ProcessIO(TContext *pContext, TResMap *pResMap)
{
  unsigned int start, stop;
  char line[500];

  while (1)
  {
    if (!GetToken(pContext, line))
      break;
    if (!isdigit(*line))
    {
      PutToken(pContext, line);
      break;
    }
    start = GetNum(line);
    if (!GetToken(pContext, line))
      break;
    if (!strcmp(line, "-"))
    {
      if (!GetToken(pContext, line))
        break;
      stop = GetNum(line);
    }
    else
    {
      PutToken(pContext, line);
      stop = start;
    }
    AddIOResMap(pResMap, start, stop); 
  }
}

static void ProcessMem(TContext *pContext, TResMap *pResMap)
{
  unsigned int start, stop;
  char line[500];

  while (1)
  {
    if (!GetToken(pContext, line))
      break;
    if (!isdigit(*line))
    {
      PutToken(pContext, line);
      break;
    }
    start = GetNum(line);
    if (!GetToken(pContext, line))
      break;
    if (!strcmp(line, "-"))
    {
      if (!GetToken(pContext, line))
        break;
      stop = GetNum(line);
    }
    else
    {
      PutToken(pContext, line);
      stop = start;
    }
    AddMemResMap(pResMap, start, stop); 
  }
}

static void ProcessIrq(TContext *pContext, TResMap *pResMap)
{
  unsigned int irq;
  char line[500];

  do
  {
    if (!GetToken(pContext, line))
      break;
    if (isdigit(*line))
    {
      irq = GetNum(line);
      AddIRQResMap(pResMap, irq);
    }
  }
  while (isdigit(*line));
  PutToken(pContext, line);
}

static void ProcessArb(TContext *pContext, TResMap *pResMap, int NoDMA)
{
  unsigned int arb;
  char line[500];

  do
  {
    if (!GetToken(pContext, line))
      break;
    if (!strcasecmp(line, "SHAREDARB"))
      GetToken(pContext, line);
    if (isdigit(*line))
    {
      arb = GetNum(line);
      AddARBResMap(pResMap, arb);
      if (!NoDMA)
        AddDMAResMap(pResMap, arb);
    }
  } while (isdigit(*line));
  PutToken(pContext, line);
}

static int poschk(TContext *pContext, uint16 *pos, int poscnt)
{
  unsigned int posind, posval, posmask;
  char line[500];

  GetToken(pContext, line); /* dump [ */
  GetToken(pContext, line);
  posind = GetNum(line) + 2;
  GetToken(pContext, line); /* dump ] */
  GetToken(pContext, line); /* dump = */
  GetToken(pContext, line); posval = GetMask(line, &posmask);
  if (posind >= poscnt)
    return -1;
  else
    return ((pos[posind] & posmask) == posval);
}

static void ProcessCard(TCardDescr *pDescr, char *prefixes)
{
  char arg[100], pref1, line[2000];
  unsigned int posid;
  FILE *adffile = NULL;
  TContext adfcontext;

  if ((pDescr->PosVals[0] != 0xff) || (pDescr->PosVals[1] != 0xff))
  {
    int cont;

    posid = (pDescr->PosVals[1] << 8) | pDescr->PosVals[0];

    if ((ListResources) || (ListOptions))
      printf("\n");
    printf("%s: ", pDescr->Name);
    if (ListIDs)
      printf("(%04X) ", posid);
    pref1= *prefixes;
    while (*prefixes)
    {
      adffile = GetADF(posid, *prefixes);
      if (adffile)
        break;
      prefixes++;
    }
    if (adffile)
    {
      TResMap CardMap;
      int NameFound, NoDMA = 0;

      InitResMap(&CardMap);
      InitContext(&adfcontext, adffile); cont = 1;
      while ((!feof(adffile)) && (cont))
      {
        if (!GetToken(&adfcontext, line))
          break;

        if (!strcasecmp(line, "ADAPTERID"))
        {
          unsigned int adfid;

          GetToken(&adfcontext, line);
          adfid = GetNum(line);
          if (posid != adfid)
          {
            printf("file name and AdapterId do not match. Oops.");
            cont = 0;
          }
        }

        else if (!strcasecmp(line, "ADAPTERNAME"))
        {
          GetToken(&adfcontext, line); printf("%s\n", line);
          NameFound = 1;
        }

        else if (!strcasecmp(line, "NODMA"))
        {
          NoDMA = 1;
        }

        else if (!strcasecmp(line, "FIXEDRESOURCES"))
        {
          int unknown = 0;

          while (!unknown)
          {
            if (!GetToken(&adfcontext, line))
              break;

            if (!strcasecmp(line, "POS"))
              poschk(&adfcontext, pDescr->PosVals, pDescr->PosCnt);
            else if (!strcasecmp(line, "IO"))
              ProcessIO(&adfcontext, &CardMap);
            else if (!strcasecmp(line, "MEM"))
              ProcessMem(&adfcontext, &CardMap);
            else if (!strcasecmp(line, "INT"))
              ProcessIrq(&adfcontext, &CardMap);
            else if (!strcasecmp(line, "ARB"))
              ProcessArb(&adfcontext, &CardMap, NoDMA);
            else if (!strcasecmp(line, "VPD"))
            {
              GetToken(&adfcontext, line); assert(!strcmp(line, "("));
              GetToken(&adfcontext, line);
              GetToken(&adfcontext, line); assert(!strcmp(line, ")"));
            }
            else
            {
              unknown = 1;
              PutToken(&adfcontext, line);
            }
          }
        }

        else if (!strcasecmp(line, "NAMEDITEM"))
        {
          char prompt[100] = "", choice[100] = "";
          int unknown;
          
          ChkNameFound(&NameFound, *prefixes, posid);

          unknown = 0;
          while (!unknown)
          {
            GetToken(&adfcontext, line);
            if (!strcasecmp(line, "PROMPT"))
              GetToken(&adfcontext, prompt);
            else if (!strcasecmp(line, "CHOICE"))
            {
              int possible = 1, unknown = 0;
              TResMap ChoiceMap;
              char thischoice[100] = "";

              InitResMap(&ChoiceMap);
              GetToken(&adfcontext, thischoice);
              if (!strcasecmp(thischoice, "VCHECK"))
                GetToken(&adfcontext, thischoice);
               
              while (!unknown)
              {
                if (!GetToken(&adfcontext, line))
                  break;

                if (!strcasecmp(line, "POS"))
                {
                  if (poschk(&adfcontext, pDescr->PosVals, pDescr->PosCnt) <= 0)
                    possible = 0;
                }
                else if (!strcasecmp(line, "IO"))
                  ProcessIO(&adfcontext, &ChoiceMap);
                else if (!strcasecmp(line, "MEM"))
                  ProcessMem(&adfcontext, &ChoiceMap);
                else if (!strcasecmp(line, "INT"))
                  ProcessIrq(&adfcontext, &ChoiceMap);
                else if (!strcasecmp(line, "ARB"))
                  ProcessArb(&adfcontext, &ChoiceMap, NoDMA);
                else
                {
                  unknown = 1;
                  PutToken(&adfcontext, line);
                }
              }
              if (possible)
              {
                strcpy(choice, thischoice);
                MergeResMap(&CardMap, &ChoiceMap);
              }
            }
            else if (!strcasecmp(line, "HELP"))
            {
              GetToken(&adfcontext, line);
            }
            else
            {
              unknown = 1;
              PutToken(&adfcontext, line);
            }
          }
          if ((*prompt) && (ListOptions))
          {
            printf("\t%s: %s\n", prompt, *choice ? choice : "unknown configuration");
          }
        }

        else if (!strcasecmp(line, "DEFSYM"))
        {
          GetToken(&adfcontext, line);
          GetToken(&adfcontext, arg);
          assert(!strcmp(arg, "="));
          GetToken(&adfcontext, arg);
          AddSymbol(line, GetNum(arg));
        }

      } /* while !feof... */
      fclose(adffile);

      ChkNameFound(&NameFound, *prefixes, posid);

      if (ListResources)
      {
        printf("\tResources: "); PrintResMap(stdout, &CardMap); printf("\n");
      }

      MergeResMap(&GlobalMap, &CardMap);
    }
    else
      printf("no ADF file '%c%04x.adf'\n", pref1, posid);
  }
}

int main(int argc, char **argv)
{
  int z, err;

  ListResources = ListOptions = ListIDs = ListSyms = err = 0;
  for (z = 1; z < argc; z++)
  {
    if (*argv[z] =='-')
    {
      switch (argv[z][1])
      {
        case 'r':
          ListResources = 1; break;
        case 's':
          ListSyms = 1; break;
        case 'o':
          ListOptions = 1; break;
        case 'i':
          ListIDs = 1; break;
        case 'a':
          ListResources = ListOptions = ListIDs = 1; break;
        default:
          err = 1;
      }
    }
  }

  if (err)
  {
    fprintf(stderr, "usage: %s [options]\n\n", *argv);
    fprintf(stderr, "options: -r : list resources occupied by cards/system\n");
    fprintf(stderr, "         -o : list device options listed in ADF files\n");
    fprintf(stderr, "         -i : list POSIDs of cards\n");
    fprintf(stderr, "         -a : list full information (-r -o -i)\n");
    fprintf(stderr, "         -s : list defined symbols\n");
    exit(1);
  }

  /* initialize overall resource map */

  InitResMap(&GlobalMap);

  /* get the POS information */

  if (!InitPosDB())
  {
    fprintf(stderr, "This machine doesn't seem to have an MCA bus.\n");
    return 1;
  }

  AddSymbol("model", ModelID);
  AddSymbol("submodel", SubmodelID);
  AddSymbol("biosrev", BIOSRev);

  /* scan through the cards: first look for planar, then for rest */

  for (z = 0; z < CardCnt; z++)
    if (!strncmp(Cards[z].Name, "Planar", 6))
    {
      ProcessCard(Cards + z, "d");
      ProcessCard(Cards + z, "p");
    }

  for (z = 0; z < CardCnt; z++)
    if (strncmp(Cards[z].Name, "Planar", 6))
      ProcessCard(Cards + z, "@_");

  if (ListResources)
  {
    printf("\n\tOverall Resources: "); PrintResMap(stdout, &GlobalMap); printf("\n");
  }

  if (ListSyms)
  {
    printf("\nSymbols: \n"); ListSymbols(stdout);
  }

  return 0;
}
