/* parser.c */
/************************************************************************/
/*                                                                      */
/*  Token Parser / Symbol list handling                                 */
/*                                                                      */
/*  changes: 2000-09-01 created                                         */
/*                                                                      */
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "parser.h"

/************************************************************************/

TSymbol *pSymList = NULL;

/************************************************************************/

/*--------------------------------------------------------------------------*/
/* symbol list */

void AddSymbol(char *pName, unsigned int value)
{
  TSymbol *pNew, *pRun, *pPrev;

  for (pRun = pSymList, pPrev = NULL; pRun; pPrev = pRun, pRun = pRun->pNext)
    if (strcasecmp(pName, pRun->pName) <= 0)
      break;

  if (!pRun)
  {
    pNew = (TSymbol*) malloc(sizeof(TSymbol));
    pNew->pNext = NULL;
    pNew->pName = strdup(pName);
    pNew->value = value;
    if (pPrev)
      pPrev->pNext = pNew;
    else
      pSymList = pNew;
  }
  else if (!strcasecmp(pName, pRun->pName))
    pRun->value = value;
  else
  {
    pNew = (TSymbol*) malloc(sizeof(TSymbol));
    pNew->pNext = pRun;
    pNew->pName = strdup(pName);
    pNew->value = value;
    if (pPrev)
      pPrev->pNext = pNew;
    else
      pSymList = pNew;
  }
}

unsigned int GetSymbol(char *pName)
{
  TSymbol *pRun;
  int cmperg;

  for (pRun = pSymList; pRun; pRun = pRun->pNext)
  {
    cmperg = strcasecmp(pName, pRun->pName);
    if (!cmperg)
      return pRun->value;
    if (cmperg < 0)
      break;
  }
  AddSymbol(pName, 0);
  return 0;
}

void ListSymbols(FILE *pFile)
{
  TSymbol *pRun;

  for (pRun = pSymList; pRun; pRun = pRun->pNext)
    printf("%-20s : %5d (0x%x)\n", pRun->pName, pRun->value, pRun->value);
}

/*--------------------------------------------------------------------------*/
/* parser help routines */

unsigned int GetNum(char *pAsc)
{
  int l = strlen(pAsc) - 1;
  char *pEnd;

  if (toupper(pAsc[l]) == 'H')
  {
    pAsc[l] = '\0';
    return strtoul(pAsc, &pEnd, 16);
  }
  else if (toupper(pAsc[l]) == 'B')
  {
    pAsc[l] = '\0';
    return strtoul(pAsc, &pEnd, 2);
  }
  else
    return strtoul(pAsc, &pEnd, 10);
}

/*--------------------------------------------------------------------------*/
/* token parsing */

#define istok(ch) ((isalnum(ch)) || (ch == '_'))

static int GetExprVal(TContext *pContext);

void InitContext(TContext *pContext, FILE *f)
{
  pContext->f = f;
  pContext->pNext = "\0";
  pContext->lasttoken[0] = '\0';
}

static int GetLine(TContext *pContext)
{
  do
  {
    if (feof(pContext->f))
      return 0;
    fgets(pContext->tmpline, sizeof(pContext->tmpline), pContext->f);
  }
  while (pContext->tmpline[0] == ';');
  pContext->pNext = pContext->tmpline;
  return 1;
}

int GetRawToken(TContext *pContext, char *pDest)
{
  char ch;

  /* tokens in buffer ? */

  if (pContext->lasttoken[0])
  {
    strcpy(pDest, pContext->lasttoken);
    pContext->lasttoken[0] = '\0';
    return 1;
  }

  /* skip leading spaces */
retry:
  for (; isspace(*(pContext->pNext)); (pContext->pNext)++)
    if (*(pContext->pNext) == '\0')
      break;

  /* input stream empty ? */

  if (*(pContext->pNext) == '\0')
  {
    if (!GetLine(pContext))
      return 0;
    goto retry;
  }

  /* copy first character: */

  ch = *(pContext->pNext++);

  /* in case this is a quotation sign, start sucking in lines until we can't go
     on any more or end mark is reached. */

  if (ch == '"')
  {
     do
     {
       ch = *(pContext->pNext++);
       switch (ch)
       {
         case '"':
         case '\r':
           break;
         case '\n':
         case '\0':
           *(pDest++) = ' ';
           if (!GetLine(pContext))
             ch = '"';
           break;
         default:
           *(pDest++) = ch;
       }
     }
     while (ch != '"');
  }

  /* normal char: go on until non-alpha character found */

  else
  {
    *(pDest++) = ch;
    if (istok(ch))
      while (istok(*(pContext->pNext)))
        *(pDest++) = *(pContext->pNext++);
  }

  /* terminate character */

  *pDest = '\0';
  return 1;
}

void PutToken(TContext *pContext, char *pTok)
{
  strcpy(pContext->lasttoken, pTok);
}

int GetToken(TContext *pContext, char *pDest)
{
  /* get a token */
redo:
  if (!GetRawToken(pContext, pDest))
    return 0;

  /* conditional ? */

  if (!strcasecmp(pDest, "IF"))
  {
    int value;

    /* get opening parenthese */

    if (!GetRawToken(pContext, pDest))
      return 0;
    assert(!strcmp(pDest, "("));

    /* get expression value */

    value = GetExprVal(pContext);
    
    /* the next token must be the opening brace */

    if (!GetRawToken(pContext, pDest))
      return 0;

    /* here we'd have to handle a single statement, which is not done so far
       since I do not know so far how a single statement can be described. */
    
    if (strcmp(pDest, "{"))
      goto redo;
    else
    {
      /* in case the condition was false, read until the corresponding
         closing brace is found.  Since this may be another IF, better
         start from beginning. If it was true, immediately restart. */

      if (!value)
      {
        int level = 0;
        do
        {
          if (!GetRawToken(pContext, pDest))
            return 0;
          if (!strcmp(pDest, "{"))
            level++;
          else if (!strcmp(pDest, "}"))
            level--;
        }
        while (level >= 0);
      }

      goto redo;
    }
  }

  /* otherwise simply return */

  else
    return 1;
}

static int GetExprVal(TContext *pContext)
{
  int SaveVal = 0, NewVal;
  char Token[200], Op[200];

  /* get tokens from the stream until we reach the closing parenthesis */

  *Op = '\0';
  do
  {
    if (!GetToken(pContext, Token))
      return 0;

    /* operator ? */

    if (!strcasecmp(Token, "eq"))
      strcpy(Op, Token);
    else if (!strcasecmp(Token, "or"))
      strcpy(Op, Token);

    /* operand ? */

    else if (strcmp(Token, ")"))
    {
      /* subexpression ? */

      if (!strcmp(Token, "("))
        NewVal = GetExprVal(pContext);

      /* simple constant? */

      else if (isdigit(*Token))
        NewVal = GetNum(Token);

      /* then variable */

      else
        NewVal = GetSymbol(Token);

      /* first part ? */

      if (!*Op)
        SaveVal = NewVal;
      else
      {
        if (!strcasecmp(Op, "eq"))
          SaveVal = (SaveVal == NewVal);
        else if (!strcasecmp(Op, "or"))
          SaveVal = (SaveVal || NewVal);
        else
          SaveVal = 0;
      }
    }
  }
  while (strcmp(Token, ")"));

  return SaveVal;
}

