/* posdb.c */
/************************************************************************/
/*                                                                      */
/*  Get POS values                                                      */
/*                                                                      */
/*  changes: 2000-09-12 created                                         */
/*           2000-09-13 added direct I/O                                */
/*                                                                      */
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <features.h>
#include <asm/io.h>
#ifdef __GLIBC__
#include <sys/io.h>
#endif

#include <string.h>
#include <ctype.h>

#include "posdb.h"

/************************************************************************/

TCardDescr Cards[20];
int CardCnt;
uint16 ModelID, SubmodelID, BIOSRev;

/************************************************************************/

static void GetIOPos(char *pName, uint16 Mask)
{
  int z;
  TCardDescr *pCard;

  if (ioperm(0x90, 0x80, 1) == -1)
  {
    fprintf(stderr, "sorry, can't get I/O privileges to gather necessary information for %s.\n", pName);
    return;
  }

  pCard = Cards + (CardCnt++);
  strcpy(pCard->Name, pName);
  outb(Mask, 0x94);
  for (z = 0; z < 6; z++)
    pCard->PosVals[z] = inb(0x100 + z);
  pCard->PosCnt = 6;
  outb(0xff, 0x94);

  ioperm(0x90, 0x100, 0);
}

static int CardPresent(char *pName)
{
  int z;

  for (z = 0; z < CardCnt; z++)
    if (!strcmp(Cards[z].Name, pName))
      return 1;

  return 0;
}

int InitPosDB(void)
{
  FILE *machfile, *posfile;
  char line[200];
  char *pSep, *pEnd, *pRes;

  /* next, try to get the machine properties */

  if (!(machfile = fopen("/proc/mca/machine", "r")))
    return 0;

  while (!feof(machfile))
  {
    fgets(line, sizeof(line), machfile);
    pSep = strchr(line, ':');
    if (pSep)
    {
      *pSep = '\0';
      if (!strcmp(line, "Model Id"))
        ModelID = strtol(pSep + 1, &pEnd, 0);
      else if (!strcmp(line, "Submodel Id"))
        SubmodelID = strtol(pSep + 1, &pEnd, 0);
      else if (!strcmp(line, "BIOS Revision"))
        BIOSRev = strtol(pSep + 1, &pEnd, 0);
    }
  }
  fclose(machfile);

  /* get via the /proc file system what is possible */

  if (!(posfile = fopen("/proc/mca/pos", "r")))
    return 0;

  CardCnt = 0;
  while (!feof(posfile))
  {
    *line = '\0';
    fgets(line, sizeof(line), posfile);
    if ((pSep = strchr(line, ':')))
    {
      TCardDescr *pCard;

      pCard = Cards + CardCnt;
      CardCnt++;

      /* split off name */

      *pSep = '\0';
      for (pEnd = pSep - 1; pEnd >= line; pEnd--)
        if (isspace(*pEnd))
          *pEnd = '\0';
        else
          break;
      strcpy(pCard->Name, line);

      /* get POS values */

      for (pSep++; *pSep; pSep++)
        if (!isspace(*pSep))
          break;
      pCard->PosCnt = 0;
      while (1)
      {
        if (!*pSep)
          break;
        pEnd = strchr(pSep, ' ');
        if (pEnd)
          *pEnd = '\0';
        pCard->PosVals[pCard->PosCnt] = strtol(pSep, &pRes, 16);
        if (*pRes != '\0')
          break;
        pCard->PosCnt++;
        if (!pEnd)
          break;
        for (pSep = pEnd + 1; *pSep; pSep++)
          if (!isspace(*pSep))
            break;
      }

      /* POS 6 and 7 are reserved fore indirect addressing */

      if (pCard->PosCnt == 8)
        pCard->PosCnt = 6;
    }
  }

  fclose(posfile);

  /* try to get missing stuff via direct I/O */

  if (!CardPresent("Planar"))
    GetIOPos("Planar", 0x7f);

  if (!CardPresent("SCSI"))
    GetIOPos("SCSI", 0xf7);

  if (!CardPresent("Video"))
    GetIOPos("Video", 0xdf);

  return 1;
}
