/* resmap.c */
/************************************************************************/
/*                                                                      */
/*  Resource map handling                                               */
/*                                                                      */
/*  changes: 2000-09-01 created                                         */
/*                                                                      */
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <string.h>

#include "resmap.h"

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif

#define NextBit(bit, ind) \
        if (bit == 0x80000000u) { bit = 1; ind++; } else bit = bit << 1

#define FirstBit(bit, ind, val) \
        { bit = 1 << (val & 31); ind = val >> 5; }

#define BitCnt(limit) (limit << 5)

/************************************************************************/

void InitResMap(TResMap *pMap)
{
  memset(pMap->iomap, 0, sizeof(pMap->iomap));
  memset(pMap->memmap, 0, sizeof(pMap->memmap));
  pMap->irqmap = pMap->dmamap = pMap->arbmap = 0;
  pMap->memlimit = pMap->iolimit = 0;
}

void MergeResMap(TResMap *pDestMap, TResMap *pSrcMap)
{
  int z;

  pDestMap->iolimit = max(pSrcMap->iolimit, pDestMap->iolimit);
  for (z = 0; z < pSrcMap->iolimit; z++)
    pDestMap->iomap[z] |= pSrcMap->iomap[z];

  pDestMap->memlimit = max(pSrcMap->memlimit, pDestMap->memlimit);
  for (z = 0; z < pSrcMap->memlimit; z++)
    pDestMap->memmap[z] |= pSrcMap->memmap[z];

  pDestMap->irqmap |= pSrcMap->irqmap;
  pDestMap->dmamap |= pSrcMap->dmamap;
  pDestMap->arbmap |= pSrcMap->arbmap;
}

void AddIOResMap(TResMap *pMap, int start, int stop)
{
  unsigned int z, cnt;
  uint32 Mask;

  FirstBit(Mask, cnt, start);
  for (z = start; z <= stop; z++)
  {
    pMap->iomap[cnt] |= Mask;
    NextBit(Mask, cnt);
    if (cnt == sizeof(pMap->iomap) / sizeof(uint32))
    {
      cnt--;
      break;
    }
  }
  if (pMap->iolimit <= cnt)
    pMap->iolimit = cnt + 1;
}

void AddMemResMap(TResMap *pMap, unsigned int start, unsigned int stop)
{
  unsigned int z, cnt, Mask;

  start = start >> 10;
  stop = stop >> 10;

  FirstBit(Mask, cnt, start);
  if (cnt >= sizeof(pMap->memmap))
    return;

  for (z = start; z <= stop; z++)
  {
    pMap->memmap[cnt] |= Mask;
    NextBit(Mask, cnt);
    if (cnt == sizeof(pMap->memmap) / sizeof(uint32))
    {
      cnt--;
      break;
    }
  }
  if (pMap->memlimit <= cnt)
    pMap->memlimit = cnt + 1;
}

static void PrintMap(FILE *pFile, int map, char *pName)
{
  int z;
  uint32 bit, first, currbit, lastbit, laststart = 0;

  bit = 1; lastbit = 0; first = 1;
  for (z = 0; z < 16; z++)
  {
    currbit = map & bit;
    if ((currbit) && (!lastbit))
    {
      fprintf(stdout, first ? "(%s " : ",", pName);
      fprintf(stdout, "%d", laststart = z);
      first = 0;
    } 
    else if ((!currbit) && (lastbit))
    {
      if (z - laststart > 1)
        fprintf(stdout, "-%d", z - 1);
    }
    lastbit = currbit;
    bit = bit << 1;
  }
  if ((lastbit) && (laststart != z - 1))
    fprintf(pFile, "-15");
  if (!first)
    fprintf(stdout, ")");
}

void PrintResMap(FILE *pFile, TResMap *pMap)
{
  unsigned int z, ind;
  uint32 bit, lastbit, currbit, first, laststart = 0;

  FirstBit(bit, ind, 0);
  lastbit = 0; first = 1;
  for (z = 0; z < BitCnt(pMap->iolimit); z++)
  {
    currbit = pMap->iomap[ind] & bit;
    if ((currbit) && (!lastbit))
    {
      fprintf(pFile, first ? "(IO " : ",");
      fprintf(pFile, "%X", laststart = z);
      first = 0;
    } 
    else if ((!currbit) && (lastbit))
    {
      if (z - laststart > 1)
        fprintf(pFile, "-%X", z - 1);
    }
    lastbit = currbit;
    NextBit(bit, ind);
  }
  if ((lastbit) && (laststart != z - 1))
    fprintf(pFile, "-%X", z - 1);
  if (!first)
    fprintf(pFile, ")");

  FirstBit(bit, ind, 0);
  lastbit = 0; first = 1;
  for (z = 0; z < BitCnt(pMap->memlimit); z++)
  {
    currbit = pMap->memmap[ind] & bit;
    if ((currbit) && (!lastbit))
    {
      fprintf(pFile, first ? "(MEM " : ",");
      fprintf(pFile, "%X", (laststart = z) << 10);
      first = 0;
    } 
    else if ((!currbit) && (lastbit))
    {
      if (z - laststart > 1)
        fprintf(pFile, "-%X", (z << 10) - 1);
    }
    lastbit = currbit;
    NextBit(bit, ind);
  }
  if ((lastbit) && (laststart != z - 1))
    fprintf(pFile, "-%X", (z << 10) - 1);
  if (!first)
    fprintf(pFile, ")");

  PrintMap(pFile, pMap->irqmap, "IRQ");
  PrintMap(pFile, pMap->dmamap, "DMA");
  PrintMap(pFile, pMap->arbmap, "ARB");
}
